/*
 * output.c - Output stRuct Content
 *
 * Written 2002 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <inttypes.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "troll.h"
#include "dwarfhelp.h"


/* ----- Helper functions -------------------------------------------------- */


static void indent(FILE *file,int level)
{
    fprintf(file,"%*s",level*4,"");
}


static unsigned long read_size(const void *p,int size)
{
    switch (size) {
	case 1:
	    return *(uint8_t *) p;
	case 2:
	    return read_16(p);
	case 4:
	    return read_32(p);
	case -1:
	    fprintf(stderr,"integer size %d not supported\n",size);
	    exit(1);
	default:
	    abort();
    }
}


/* ----- Type output ------------------------------------------------------- */


static void output_enum_members_type(FILE *file,const struct enum_member *m,
  int level)
{
    while (m) {
	indent(file,level);
	fprintf(file,"%s = %d\n",m->name,m->value);
	m = m->next;
    }
}


static void do_output_type(FILE *file,const struct troll *t,int level,
  int terse);


static void output_struct_members_type(FILE *file,const struct struct_member *m,
  int level)
{
    while (m) {
	indent(file,level);
	fprintf(file,".%s <@%lu>\n",m->name,m->offset);
	do_output_type(file,m->type,level+1,0);
	m = m->next;
    }
}


static void do_output_type(FILE *file,const struct troll *t,int level,
  int terse)
{
    indent(file,level);
    if (!t) {
	fprintf(file,"void\n");
	return;
    }
    switch (t->type) {
	case troll_uint:
	    switch (t->size) {
		case 1:
		    fprintf(file,"unsigned char\n");
		    break;
		case 2:
		    fprintf(file,"uint16_t\n");
		    break;
		case 4:
		    fprintf(file,"uint32_t\n");
		    break;
		case -1:
		    fprintf(file,"unsigned int <%d>\n",t->size);
		    break;
		default:
		    abort();
	    }
	    break;
	case troll_sint:
	    switch (t->size) {
		case 1:
		    fprintf(file,"char\n");
		    break;
		case 2:
		    fprintf(file,"int16_t\n");
		    break;
		case 4:
		    fprintf(file,"int32_t\n");
		    break;
		case -1:
		    fprintf(file,"int <%d>\n",t->size);
		    break;
		default:
		    abort();
	    }
	    break;
	case troll_float:
	    fprintf(file,"float <%d>\n",t->size);
	    break;
	case troll_enum:
	    fprintf(file,"enum %s <%d>\n",t->name ? t->name : "",t->size);
	    output_enum_members_type(file,t->u._enum.members,level+1);
	    break;
	case troll_const:
	    fprintf(file,"const <%d>\n",t->size);
	    do_output_type(file,t->u.modifier.type,level+1,terse);
	    break;
	case troll_ptr:
	    fprintf(file,"* <%d>\n",t->size);
	    do_output_type(file,t->u.modifier.type,level+1,1);
	    break;
	case troll_volatile:
	    fprintf(file,"volatile <%d>\n",t->size);
	    do_output_type(file,t->u.modifier.type,level+1,terse);
	    break;
	case troll_fn:
	    fprintf(file,"fn() <%d>\n",t->size);
	    do_output_type(file,t->u.fn.type,level+1,1);
	    break;
	case troll_array:
	    fprintf(file,"[%u] <%d, *%d>\n",t->u.array.elements,t->size,
	      t->u.array.stride_size);
	    do_output_type(file,t->u.array.type,level+1,terse);
	    break;
	case troll_struct:
	    fprintf(file,"struct %s <%d>\n",t->name ? t->name : "",t->size);
	    if (!terse)
		output_struct_members_type(file,t->u._struct.members,level+1);
	    break;
	case troll_union:
	    fprintf(file,"union %s <%d>\n",t->name ? t->name : "",t->size);
	    if (!terse)
		output_struct_members_type(file,t->u._struct.members,level+1);
	    break;
	case troll_typedef:
	    fprintf(file,"typedef %s <%d>\n",t->name,t->size);
	    do_output_type(file,t->u._typedef.type,level+1,terse);
	    break;
	default:
	    abort();
    }
}


void output_type(FILE *file,const struct troll *troll)
{
    do_output_type(file,troll,0,0);
}


/* ----- Guesstimate size of type ------------------------------------------ */


int output_sizeof(const struct troll *t)
{
    switch (t->type) {
	const struct struct_member *m;
	int size,max;

	case troll_uint:
	case troll_sint:
	case troll_float:
	    assert(t->size != -1);
	    return t->size;
	case troll_enum:
	    return t->size == -1 ? 4 : t->size;
	case troll_const:
	case troll_volatile:
	    if (t->size != -1) return t->size;
	    return output_sizeof(t->u.modifier.type);
	case troll_ptr:
	    return 4;
	case troll_array:
	    if (t->size != -1) return t->size;
	    if (t->u.array.stride_size == -1)
		return t->u.array.elements*t->u.array.stride_size;
	    return t->u.array.elements*output_sizeof(t->u.array.type);
	case troll_struct:
	    if (t->size != -1) return t->size;
	    size = 0;
	    for (m = t->u._struct.members; m; m = m->next)
		size += output_sizeof(m->type);
	    return size;
	case troll_union:
	    if (t->size != -1) return t->size;
	    max = 0;
	    for (m = t->u._struct.members; m; m = m->next) {
		size = output_sizeof(m->type);
		if (size > max) max = size;
	    }
	    return max;
	case troll_typedef:
	    return output_sizeof(t->u._typedef.type);
	default:
	    abort();
    }
}


/* ----- Data output ------------------------------------------------------- */


static void do_output_data(FILE *file,const struct troll *t,
  const void *data,unsigned long offset,int level);


static void output_enum_members_data(FILE *file,
  const struct enum_member *m,unsigned long value)
{
    while (m) {
	if (m->value == value) {
	    fprintf(file,"%s\n",m->name);
	    return;
	}
	m = m->next;
    }
    fprintf(file,"%lu\n",value);
}


static void output_struct_members_data(FILE *file,
  const struct struct_member *m,const void *data,unsigned long offset,
  int level)
{
    while (m) {
	indent(file,level);
	fprintf(file,"%s = ",m->name);
	do_output_data(file,m->type,data,offset+m->offset,level);
	m = m->next;
    }
}


static void do_output_data(FILE *file,const struct troll *t,
  const void *data,unsigned long offset,int level)
{
    const void *p;
    int i;

    p = data+offset;
    switch (t->type) {
	case troll_uint:
	    fprintf(file,"%lu\n",read_size(p,t->size));
	    break;
	case troll_sint:
	    switch (t->size) {
		long tmp;

		case 1:
		    fprintf(file,"%d\n",*(int8_t *) p);
		    break;
		case 2:
		    tmp = read_16(p);
		    if (tmp & 0x8000) tmp |= ~0 & ~0xffff;
		    fprintf(file,"%ld\n",tmp);
		    break;
		case 4:
		    tmp = read_32(p);
		    if (tmp & 0x80000000) tmp |= ~0 & ~0xffffffff;
		    fprintf(file,"%ld\n",tmp);
		    break;
		case -1:
		    fprintf(stderr,"integer size %d not supported\n",t->size);
		    exit(1);
		    break;
		default:
		    abort();
	    }
	    break;
	case troll_float:
	    fprintf(stderr,"we don't support floats yet, sorry\n");
	    exit(1);
	case troll_enum:
	    output_enum_members_data(file,t->u._enum.members,
	      read_size(p,t->size == -1 ? 4 : t->size));
	    break;
	case troll_ptr:
	    fprintf(file,"0x%lx\n",(unsigned long) read_32(p));
	    break;
	case troll_const:
	case troll_volatile:
	    do_output_data(file,t->u.modifier.type,data,offset,level);
	    break;
	case troll_array:
	    fprintf(file,"{\n");
	    for (i = 0; i < t->u.array.elements; i++) {
		indent(file,level+1);
		fprintf(file,"[%d] = ",i);
		do_output_data(file,t->u.array.type,data,offset,level+1);
		if (t->u.array.stride_size == -1)
		    offset += output_sizeof(t->u.array.type);
		else offset += t->u.array.stride_size;
	    }
	    indent(file,level);
	    fprintf(file,"}\n");
	    break;
	case troll_struct:
	    fprintf(file,"{\n");
	    output_struct_members_data(file,t->u._struct.members,data,offset,
	      level+1);
	    indent(file,level);
	    fprintf(file,"}\n");
	    break;
	case troll_union:
	    fprintf(file,"{\n");
	    output_struct_members_data(file,t->u._struct.members,data,offset,
	      level+1);
	    indent(file,level);
	    fprintf(file,"}\n");
	    break;
	case troll_typedef:
	    do_output_data(file,t->u._typedef.type,data,offset,level);
	    break;
	default:
	    abort();
    }
}


void output_data(FILE *file,const struct troll *troll,const void *data)
{
    do_output_data(file,troll,data,0,0);
}


/* ----- Variable lookup --------------------------------------------------- */


static const struct troll *lookup_var(const struct troll *t,va_list ap,
  unsigned long *offset)
{
    while (1) {
	switch (t->type) {
	    const struct struct_member *m;
	    const char *name;
	    int num;

	    case troll_uint:
	    case troll_sint:
	    case troll_enum:
	    case troll_ptr:
		return t;
	    case troll_const:
	    case troll_volatile:
		t = t->u.modifier.type;
		break;
	    case troll_typedef:
		t = t->u._typedef.type;
		break;
	    case troll_array:
		num = va_arg(ap,int);
		*offset += num*(t->u.array.stride_size != -1 ?
		  t->u.array.stride_size : output_sizeof(t->u.array.type));
		t = t->u.array.type;
		break;
	    case troll_struct:
	    case troll_union:
		name = va_arg(ap,const char *);
		for (m = t->u._struct.members; m; m = m->next)
		     if (!strcmp(m->name,name)) break;
		if (!m) {
		     fprintf(stderr,"no member \"%s\"\n",name);
		     exit(1);
		}
		*offset += m->offset;
		t = m->type;
		break;
	    default:
		abort();
	}
    }
}


unsigned long output_lookup_unsigned(const struct troll *t,const void *data,
  ...)
{
    unsigned long offset = 0;
    va_list ap;

    va_start(ap,data);
    t = lookup_var(t,ap,&offset);
    va_end(ap);
    return read_size(data+offset,t->size);
}


long output_lookup_signed(const struct troll *t,const void *data,...)
{
    unsigned long offset = 0;
    va_list ap;

    va_start(ap,data);
    t = lookup_var(t,ap,&offset);
    va_end(ap);
    return read_size(data+offset,t->size); /* need to sign-extend ... */
}
