/*
 * troll.c - Types Read frOm eLf fiLe
 *
 * Written 2002 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "dwarf.h"
#include "troll.h"


#define TROLL_HASH 4111
static struct troll *troll_hash[TROLL_HASH];


unsigned long loc_comp(const void *expr,int length,unsigned long base);


static const struct troll last_child; /* end marker (returned by read_type) */


static const struct troll *read_type(const void **p);


/* ----- Type lookup ------------------------------------------------------- */


const struct troll *troll_lookup(enum troll_type type,const char *name,
  int size,int skip_n)
{
    int i;
    const struct troll *walk;

    for (i = 0; i < TROLL_HASH; i++)
	for (walk = troll_hash[i]; walk; walk = walk->next) {
	    if (walk->type != type) continue;
	    if (name && (!walk->name || strcmp(name,walk->name))) continue;
	    if (size != -1 && walk->size != -1 && size != walk->size) continue;
	    if (!skip_n--) return walk;
	}
    return NULL;

}


/* ----- Helper functions -------------------------------------------------- */


static const struct dwarf_attr *require(enum DW_AT attr_id)
{
    const struct dwarf_attr *attr;

    attr = dwarf_attr(attr_id);
    if (attr) return attr;
    fprintf(stderr,"required attribute %u not found\n",attr_id);
    exit(1);
}


static void skip_children(const void **p)
{
    if (dwarf_children)
	while (dwarf_tag(p))
	    skip_children(p);
}


/* ----- Allocation and hashing -------------------------------------------- */


static struct troll *get_troll(unsigned long def)
{
    struct troll *troll;
    int hash;

    if (!def) return NULL;
    hash = def % TROLL_HASH;
    for (troll = troll_hash[def % TROLL_HASH]; troll; troll = troll->next)
	if (troll->def == def) break;
    return troll;
}


static struct troll *new_troll(enum troll_type type,unsigned long def)
{
    struct troll *troll;
    const struct dwarf_attr *attr;
    int hash;

    if (get_troll(def)) {
	fprintf(stderr,"new_troll: type %lu already exists\n",def);
	exit(1);
    }
    troll = malloc(sizeof(struct troll));
    if (!troll) {
	perror("malloc");
	exit(1);
    }
    troll->type = type;
    attr = dwarf_attr(DW_AT_name);
    troll->name = attr ? dwarf_attr_string(attr) : NULL;
    attr = dwarf_attr(DW_AT_byte_size);
    troll->size = attr ? dwarf_attr_uconst(attr) : -1;
    troll->def = def;
    hash = def % TROLL_HASH;
    troll->next = troll_hash[hash];
    troll_hash[hash] = troll;
    return troll;
}


/* ----- Resolve types ----------------------------------------------------- */


/*
 * Warning: resolve_type will clobber the attribute cache, so call it only
 * after all other attributes have been retrieved.
 */


static const struct troll *resolve_type(void)
{
    const struct dwarf_attr *attr;
    const void *ref;
    const struct troll *troll;

    attr = dwarf_attr(DW_AT_type);
    if (!attr) return NULL; /* void */
    ref = dwarf_attr_reference(attr);
    troll = get_troll((unsigned long) ref);
    if (troll) return troll;
    troll = read_type(&ref);
    if (troll) return troll;
    fprintf(stderr,"failed to resolve type at %p\n",ref);
    exit(1);
}


/* ----- Type parsers ------------------------------------------------------ */


static const struct troll *array_type(const void **p,unsigned long def)
{
    struct troll *troll;
    const struct troll **last_type = (const struct troll **) &troll;
    const struct dwarf_attr *attr;
    int stride_size = -1;

    assert(dwarf_children);
    attr = dwarf_attr(DW_AT_stride_size);
    if (attr) stride_size = dwarf_attr_uconst(attr);
    *last_type = resolve_type();
    while (1) {
	enum DW_TAG tag;
	struct troll *new;

	tag = dwarf_tag(p);
	if (!tag) break;
	new = new_troll(troll_array,def);
	new->u.array.stride_size = stride_size;
	stride_size = -1;
	new->u.array.type = *last_type;
	*last_type = new;
	last_type = &new->u.array.type;
	switch (tag) {
	    case DW_TAG_subrange_type:
		attr = dwarf_attr(DW_AT_upper_bound);
		if (attr) {
		    new->u.array.elements = dwarf_attr_uconst(attr)+1;
		    break;
		}
		attr = dwarf_attr(DW_AT_count);
		if (attr) {
		    new->u.array.elements = dwarf_attr_uconst(attr);
		    break;
		}
		new->u.array.elements = -1; /* unspecified */
		break;
	    case DW_TAG_enumeration_type:
		abort();
	    default:
		abort();
	}
	def = 0;
    }
    return troll;
}


static const struct troll *struct_type(enum troll_type kind,const void **p,
  unsigned long def)
{
    struct troll *troll;
    struct struct_member **last;

    troll = new_troll(kind,def);
    troll->u._struct.members = NULL;
    if (!dwarf_children) return troll;
    last = &troll->u._struct.members;
    while (1) {
	enum DW_TAG tag;
	const struct dwarf_attr *attr;

	tag = dwarf_tag(p);
	if (!tag) break;
	assert(tag == DW_TAG_member);
	*last = malloc(sizeof(struct struct_member));
	if (!*last) {
	    perror("malloc");
	    exit(1);
	}
	(*last)->name = dwarf_attr_string(require(DW_AT_name));
	attr = dwarf_attr(DW_AT_byte_size);
	(*last)->size = attr ? dwarf_attr_uconst(attr) : -1;
	attr = dwarf_attr(DW_AT_data_member_location);
	if (!attr) (*last)->offset = 0;
	else {
	    const void *expr;
	    int length;

	    expr = dwarf_attr_block(attr,&length);
	    /*
	     * This isn't entirely correct, because the location expression
	     * might produce different results if the base is non-zero
	     * (e.g. for alignment). @@@
	     *
	     * Cleaner solution: pass the current offset along while
	     * resolving nested types.
	     */
	    (*last)->offset = loc_comp(expr,length,0);
	}
	(*last)->type = resolve_type();
	last = &(*last)->next;
	*last = NULL;
    }
    return troll;
}


static const struct troll *enum_type(const void **p,unsigned long def)
{
    struct troll *troll;
    struct enum_member **last;

    troll = new_troll(troll_enum,def);
    troll->u._enum.members = NULL;
    if (!dwarf_children) return troll;
    last = &troll->u._enum.members;
    while (1) {
	enum DW_TAG tag;

	tag = dwarf_tag(p);
	if (!tag) break;
	assert(tag == DW_TAG_enumerator);
	*last = malloc(sizeof(struct enum_member));
	if (!*last) {
	    perror("malloc");
	    exit(1);
	}
	(*last)->name = dwarf_attr_string(require(DW_AT_name));
	(*last)->value = dwarf_attr_sconst(require(DW_AT_const_value));
	last = &(*last)->next;
	*last = NULL;
    }
    return troll;
}


static const struct troll *type_modifier(enum troll_type type,
  unsigned long def)
{
    struct troll *troll;

    troll = new_troll(type,def);
    troll->u.modifier.type = resolve_type();
    return troll;
}


static const struct troll *subroutine_type(unsigned long def)
{
    struct troll *troll;

    troll = new_troll(troll_fn,def);
    troll->u.fn.type = resolve_type();
    return troll;
}


static const struct troll *typedef_type(unsigned long def)
{
    struct troll *troll;

    troll = new_troll(troll_typedef,def);
    troll->u._typedef.type = resolve_type();
    return troll;
}


static const struct troll *base_type(unsigned long def)
{
    enum DW_ATE enc;
    struct troll *troll;

    enc = dwarf_attr_uconst(require(DW_AT_encoding));
    switch (enc) {
	case DW_ATE_signed:
	case DW_ATE_signed_char:
	    troll = new_troll(troll_sint,def);
	    break;
	case DW_ATE_unsigned:
	case DW_ATE_unsigned_char:
	    troll = new_troll(troll_uint,def);
	    break;
	case DW_ATE_float:
	    troll = new_troll(troll_float,def);
	    break;
	default:
	    fprintf(stderr,"unknown encoding %u\n",enc);
	    exit(1);
    }
    return troll;
}


/* ----- Main loop --------------------------------------------------------- */


static const struct troll *read_type(const void **p)
{
    unsigned long def;
    const struct troll *troll;

    def = (unsigned long) *p;
    troll = get_troll(def);
    if (troll) {
	dwarf_tag(p);
	skip_children(p);
	return troll;
    }
    switch (dwarf_tag(p)) {
	case DW_TAG_array_type:
	    return array_type(p,def);
	case DW_TAG_enumeration_type:
	    return enum_type(p,def);
	case DW_TAG_structure_type:
	    return struct_type(troll_struct,p,def);
	case DW_TAG_subroutine_type:
	    return subroutine_type(def);
	case DW_TAG_typedef:
	    return typedef_type(def);
	case DW_TAG_union_type:
	    return struct_type(troll_union,p,def);
	case DW_TAG_base_type:
	    return base_type(def);
	case DW_TAG_pointer_type:
	    return type_modifier(troll_ptr,def);
	case DW_TAG_const_type:
	    return type_modifier(troll_const,def);
	case DW_TAG_volatile_type:
	    return type_modifier(troll_volatile,def);
	case 0:
	    return &last_child;
	default:
	    return NULL;
    }
}


void read_types(const void *unit)
{
    while (dwarf_more_tags(unit))
	read_type(&unit);
}
